# coding: utf-8

"""
    almdrlib Exceptions
"""


class AlmdrlibException(Exception):
    """The base exception class for all almdrlib exceptions"""
    pass


class AlmdrlibTypeError(AlmdrlibException, TypeError):
    def __init__(self, msg):
        super().__init__(msg)


class AlmdrlibValueError(AlmdrlibException, ValueError):
    def __init__(self, msg):
        super().__init__(msg)


class AlmdrlibKeyError(AlmdrlibException, KeyError):
    def __init__(self, msg):
        super().__init__(msg)



# -*- coding: utf-8 -*-

from enum import Enum

"""
    alertlogic.region
    ~~~~~~~~~~~~~~
    alertlogic region management
"""
"""
List of known regions
"""
ENDPOINTS = {
    "production": "https://api.global-services.global.alertlogic.com",
    "integration": "https://api.global-integration.product.dev.alertlogic.com"
}


class NoValue(Enum):
    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)


class Residency(NoValue):
    @staticmethod
    def list_residencies():
        return ['default', 'us', 'emea']

    DEFAULT = 'default'
    US = 'us'
    EMEA = 'emea'


class EndpointType(NoValue):
    API = 'api'
    UI = 'ui'


class Region():
    """
    Abstracts an alertlogic region.
    For now it only represents the api endpoint url
    """

    def __init__(self):
        pass

    @staticmethod
    def list_endpoints():
        return ENDPOINTS.keys()

    @staticmethod
    def get_global_endpoint(endpoint):
        return ENDPOINTS.get(endpoint, endpoint)

    @staticmethod
    def get_region_from_location(location):
        return location.split('-')[1]

    @staticmethod
    def get_endpoint_url(url, service_name, account_id, residency="default"):
        return "{}/endpoints/v1/{}/residency/{}/services/{}/endpoint".format(
                url,
                account_id,
                residency,
                service_name)



# -*- coding: utf-8 -*-

"""
    almdrlib.client
    ~~~~~~~~~~~~~~~
    almdrlib OpenAPI v3 dynamic client builder
"""

import abc
# import typing
# import inspect
import logging
import json
import jsonschema
from jsonschema.validators import validator_for
import alsdkdefs

from almdrlib.exceptions import AlmdrlibValueError
from almdrlib.config import Config

from alsdkdefs import OpenAPIKeyWord

logger = logging.getLogger(__name__)


class Server(object):
    def __init__(self, service_name, spec,
                 session=None, variables=None):
        self._service_name = service_name
        self._spec = spec
        self._session = session
        self._url = spec.get(OpenAPIKeyWord.URL)
        self._description = spec.get(OpenAPIKeyWord.DESCRIPTION)

        if variables:
            self.variables = variables
        elif OpenAPIKeyWord.VARIABLES in spec:
            self.variables = dict(
                    (k, v.get(OpenAPIKeyWord.DEFAULT))
                    for k, v in spec[OpenAPIKeyWord.VARIABLES].items())
        else:
            self.variables = variables

        if spec.get(OpenAPIKeyWord.X_ALERTLOGIC_SESSION_ENDPOINT) and \
                self._session:
            self._url = self._session.get_url(self._service_name)

        logger.debug(f"Server initialized using '{self._url}' URL " +
                     f"for '{self._service_name}' service.")

    @property
    def url(self):
        if self.variables:
            return self._url.format(**self.variables)
        else:
            return self._url

    @property
    def spec(self):
        return self._spec

    def set_url(self, url):
        self._url = url


class OperationResponse(object):
    _response_schema = {}

    def __init__(self, schema, session=None):
        for r_code, r_schema in schema.items():
            if r_code[0] == '2':
                self._add_response(r_schema.pop(OpenAPIKeyWord.CONTENT, None))

        self._schema = schema

    def _add_response(self, content):
        if not content:
            return

        # While there could be more than one content type,
        # We support only the first content type.
        content_type, content_type_schema = next(iter(content.items()))
        if content_type not in OpenAPIKeyWord.JSON_CONTENT_TYPES:
            logger.warn(
                    f"{content_type} content type is unsupported." +
                    f"Only {OpenAPIKeyWord.JSON_CONTENT_TYPES} " +
                    "content types are supported"
            )
        if not content_type_schema:
            return

        self._response_schema = content_type_schema.get(OpenAPIKeyWord.SCHEMA)

    @property
    def schema(self):
        return self._response_schema

    @property
    def exceptions(self):
        return {}


class RequestBodyParameter(object):
    def __init__(self, name, schema, required=False, session=None):
        self._name = name
        self._schema = schema
        self._required = required
        self._session = session

        validator_cls = validator_for(schema)
        validator_cls.check_schema(schema)
        self._validator = validator_cls(schema)

    @abc.abstractmethod
    def serialize(self, value, header=[]):
        """ Derived classes handle serialization """
        return

    def validate(self, data):
        try:
            self._validator.validate(data)
        except jsonschema.exceptions.ValidationError as e:
            raise AlmdrlibValueError(
                    f"{e.message}. Schema: {json.dumps(e.schema)}")

    @property
    def name(self):
        return self._name

    @property
    def required(self):
        return self._required

    @property
    def schema(self):
        return {
            self._name: self._schema
        }


class RequestBodySchemaParameter(RequestBodyParameter):
    def __init__(self, name, schema, required=False, session=None):
        super().__init__(name, schema, required, session)

    def serialize(self, kwargs, header=[]):
        data = kwargs.pop(self.name, {})
        self.validate(data)
        kwargs['data'] = json.dumps(data)


class RequestBodySimpleParameter(RequestBodyParameter):
    def __init__(self, name, schema, required=False, session=None):
        super().__init__(name, schema, required, session)
        self._format = schema.get(OpenAPIKeyWord.FORMAT)

    def serialize(self, kwargs, header=None):
        data = kwargs.pop(self.name, "")
        if self._format == 'binary':
            kwargs['data'] = data.encode()
        else:
            kwargs['data'] = data


class RequestBodyObjectParameter(RequestBodyParameter):
    def __init__(self,
                 name,
                 schema,
                 al_schema={},
                 required=False,
                 session=None):
        super().__init__(
                name,
                _normalize_schema(name,
                                  schema,
                                  required),
                required,
                session
            )

        self._encoding = al_schema.pop(OpenAPIKeyWord.ENCODING, None)
        self._explode = \
            self._encoding and \
            self._encoding.get(OpenAPIKeyWord.EXPLODE, False)

        #
        # Get request body object properties
        #
        self._properties = self._schema.get(OpenAPIKeyWord.PROPERTIES, None)
        self._required_properties = self._schema.get(
                                            OpenAPIKeyWord.REQUIRED, [])

        if self._properties:
            for name in self._required_properties:
                self._properties[name].update(
                        {'x-alertlogic-required': True})

        logger.debug(
                "Initialized body parameter. "
                f"Name: {self.name}. "
                "Properties: "
                f"{self._properties and self._properties.keys() or self.name}"
                "Required Properties: "
                f"{self._required_properties}"
            )

    def serialize(self, kwargs, headers=None):
        if not all(name in kwargs for name in self._required_properties):
            raise AlmdrlibValueError(
                f"'{self._required_properties}' parameters are required. " +
                f"'{kwargs}' were provided.")

        result = {
                    k: kwargs.pop(k)
                    for k in self._properties.keys() if k in kwargs
                }

        if self.required and not bool(result):
            raise AlmdrlibValueError(
                "At least one the " +
                f"{self._properties.keys()} parameters must be specified."
            )

        # Validate provided payload against the schema
        self.validate(result)

        if self._explode:
            kwargs['data'] = json.dumps(result.pop(self.name))
        else:
            kwargs['data'] = json.dumps(result)

    @property
    def schema(self):
        return self._properties


class RequestBody(object):
    def __init__(self, required=False, description=None, session=None):
        self._parameters = {}
        self._required = required
        self._description = description
        self._session = session
        self._content_types = {}
        self._content = {}
        self._default_content_type = None

    def add_content(self, content_type, schema, al_schema):
        if not schema:
            return

        datatype = schema.get(OpenAPIKeyWord.TYPE)
        name = al_schema.get(OpenAPIKeyWord.NAME, OpenAPIKeyWord.DATA)
        if datatype == OpenAPIKeyWord.OBJECT:
            parameter = RequestBodyObjectParameter(
                            name=name,
                            schema=schema,
                            al_schema=al_schema,
                            required=self._required,
                            session=self._session
                        )
        elif datatype in OpenAPIKeyWord.SIMPLE_DATA_TYPES:
            parameter = RequestBodySimpleParameter(
                            name=name,
                            schema=schema,
                            required=self._required,
                            session=self._session
                        )
        else:
            parameter = RequestBodySchemaParameter(
                            name=name,
                            schema=schema,
                            required=self._required,
                            session=self._session
                        )

        self._content[content_type] = parameter

        if name in self._parameters:
            self._parameters[name].update({content_type: parameter.schema})
        else:
            self._parameters[name] = {content_type: parameter.schema}

    def serialize(self, headers, kwargs):
        #
        # Get content parameters.
        #
        if OpenAPIKeyWord.CONTENT_TYPE_PARAM in headers:
            content_type = headers[OpenAPIKeyWord.CONTENT_TYPE_PARAM]
            payloadBodyParam = self._content[content_type]
        elif len(self._content) == 1:
            content_type, payloadBodyParam = next(iter(self._content.items()))
            headers[OpenAPIKeyWord.CONTENT_TYPE_PARAM] = content_type
        else:
            raise AlmdrlibValueError(
                f"'{OpenAPIKeyWord.CONTENT_TYPE_PYTHON_PARAM}'" +
                "parameter is required.")

        payloadBodyParam.serialize(kwargs, headers)

    def get_schema(self):
        if len(self._content) == 1:
            payloadBodyParam = next(iter(self._content.values()))
            return {OpenAPIKeyWord.PROPERTIES: payloadBodyParam.schema}

        # Request body supports has multiple content types
        properties = dict()
        for name, schema in self._parameters.items():
            properties[name] = {
                    'content': {
                        content_type: parameter.get(name)
                        for content_type, parameter in schema.items()
                    }
                }

        return {
                OpenAPIKeyWord.PROPERTIES: properties,
                'x-alertlogic-payload-content': self._get_content_schema()
            }

    def _get_content_schema(self):
        return {
            name: list(property.schema.keys())
            for name, property in self._content.items()
        }

    def _get_required_properties(self, content):
        return [property.name
                for property in content.values() if property.required]


class PathParameter(object):
    def __init__(self, spec={}, session=None):
        # TODO: Rework PathParameter to work based on the saved spec
        self._in = spec[OpenAPIKeyWord.IN]
        self._init_name(spec[OpenAPIKeyWord.NAME])
        self._required = spec.get(OpenAPIKeyWord.REQUIRED, False)
        self._description = spec.get(OpenAPIKeyWord.DESCRIPTION, "")
        self._datatype = get_dict_value(
                            spec,
                            [OpenAPIKeyWord.SCHEMA, OpenAPIKeyWord.TYPE],
                            OpenAPIKeyWord.STRING)
        self._style = spec.get(OpenAPIKeyWord.STYLE,
                               self.default_style(self._in))
        self._explode = spec.get(OpenAPIKeyWord.EXPLODE,
                                 self.default_explode(self._style))
        self._spec = spec
        self._session = session
        self._default = None

    def _init_name(self, name):
        self._name = name.replace('-', '_')
        self._schema_name = name

    @property
    def name(self):
        return self._name

    @property
    def schema_name(self):
        return self._schema_name

    @property
    def required(self):
        return self._required

    @property
    def description(self):
        return self._description

    @property
    def datatype(self):
        return self._datatype

    @property
    def default(self):
        if self._default is None:
            self._default = self._session.get_default(self._name)
        return self._default

    @property
    def schema(self):
        result = {}
        for name, value in self._spec.items():
            if OpenAPIKeyWord.SCHEMA == name:
                result.update({k: v for k, v in value.items()})
            elif OpenAPIKeyWord.NAME == name:
                continue
            else:
                result[name] = value

        return result

    def serialize(self, path_params, query_params, headers, cookies, kwargs):
        if self._name not in kwargs and not self.default:
            if self._required:
                raise ValueError(f"'{self._name}' is required")
            return

        raw_value = kwargs.pop(self._name, self.default)

        value = serialize_value(
                self._datatype,
                raw_value)

        if self._in == OpenAPIKeyWord.PATH:
            path_params[self.schema_name] = value
        elif self._in == OpenAPIKeyWord.QUERY:
            new_query_params = self.serialize_query_parameter(self._style,
                                                              self._explode,
                                                              self._name,
                                                              self._datatype,
                                                              raw_value)
            query_params.update(new_query_params)
        elif self._in == OpenAPIKeyWord.HEADER:
            headers[self.schema_name] = value
        elif self._in == OpenAPIKeyWord.COOKIE:
            cookies[self.schema_name] = value

        return True

    @classmethod
    def default_style(cls, parameter_in):
        if parameter_in == OpenAPIKeyWord.QUERY:
            return OpenAPIKeyWord.PARAMETER_STYLE_FORM
        elif parameter_in == OpenAPIKeyWord.PATH:
            return OpenAPIKeyWord.PARAMETER_STYLE_SIMPLE
        elif parameter_in == OpenAPIKeyWord.HEADER:
            return OpenAPIKeyWord.PARAMETER_STYLE_SIMPLE
        elif parameter_in == OpenAPIKeyWord.COOKIE:
            return OpenAPIKeyWord.PARAMETER_STYLE_FORM
        else:
            return OpenAPIKeyWord.PARAMETER_STYLE_SIMPLE

    @classmethod
    def default_explode(cls, style):
        if style == OpenAPIKeyWord.PARAMETER_STYLE_FORM:
            return True
        else:
            return False

    @classmethod
    def serialize_query_parameter(cls, style, explode, name, datatype, value):
        # Implements partial query parameter serialization using rules from:
        # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md#style-examples
        # https://swagger.io/docs/specification/serialization/#query
        # TODO: Serialize deepObject style
        valid_styles = [OpenAPIKeyWord.PARAMETER_STYLE_FORM,
                        OpenAPIKeyWord.PARAMETER_STYLE_SPACE_DELIMITED,
                        OpenAPIKeyWord.PARAMETER_STYLE_PIPE_DELIMITED]
        if style not in valid_styles:
            raise ValueError(f"{name} query parameter has invalid style: "
                             f"{style}")
            return

        if (datatype == OpenAPIKeyWord.OBJECT and
                style == OpenAPIKeyWord.PARAMETER_STYLE_FORM):
            if explode:
                return value
            else:
                serialized_pairs = [f'{a},{b}' for (a, b) in value.items()]
                return {name: ",".join(serialized_pairs)}
        elif datatype == OpenAPIKeyWord.ARRAY:
            if explode:
                return {name: value}
            else:
                if style == OpenAPIKeyWord.PARAMETER_STYLE_SPACE_DELIMITED:
                    delimiter = " "
                elif style == OpenAPIKeyWord.PARAMETER_STYLE_PIPE_DELIMITED:
                    delimiter = "|"
                else:
                    # Implicitly 'form' style
                    delimiter = ","
                return {name: delimiter.join(value)}
        elif datatype == OpenAPIKeyWord.BOOLEAN:
            return {name: str(value).lower()}
        else:
            return {name: value}


class Operation(object):
    _internal_param_prefix = "_"
    _call = None

    def __init__(self,
                 path,
                 params,
                 summary,
                 description,
                 method, spec,
                 body,
                 response,
                 session=None,
                 server=None):
        self._path = path
        self._params = params
        self._summary = summary
        self._description = description
        self._method = method
        self._spec = spec
        self._body = body
        self._response = response
        self._session = session
        self._server = server
        self._operation_id = self._spec[OpenAPIKeyWord.OPERATION_ID]

        logger.debug(f"Initilized {self._operation_id} operation.")

    @property
    def spec(self):
        return self._spec

    @property
    def operation_id(self):
        return self._operation_id

    @property
    def method(self):
        return self._method

    @property
    def description(self):
        return self._description

    @property
    def path(self):
        return self._path

    @property
    def params(self):
        return self._params

    @property
    def body(self):
        return self._body

    def url(self, **kwargs):
        return self._server.url + self._path.format(**kwargs)

    def get_schema(self):
        result = {
            OpenAPIKeyWord.OPERATION_ID: self.operation_id,
            OpenAPIKeyWord.DESCRIPTION: self.description
        }
        params_schema = {}
        for param in self.params:
            params_schema.update({param.name: param.schema})

        if self.body:
            schema = self.body.get_schema()
            params_schema.update(schema.get(OpenAPIKeyWord.PROPERTIES))
            payload_content = schema.get('x-alertlogic-payload-content')
            if payload_content:
                params_schema['content_type'].update(
                        {'x-alertlogic-payload-content': payload_content}
                )

            # if OpenAPIKeyWord.CONTENT in schema:
                # result.update(schema)
            # if 'x-alertlogic-payload' in schema:
            #     params_schema['content_type'].update(schema)
            # else:
            #     params_schema.update(schema.get(OpenAPIKeyWord.PROPERTIES))

        result.update({
            OpenAPIKeyWord.PARAMETERS: dict(sorted(params_schema.items()))
        })

        result.update({
            OpenAPIKeyWord.RESPONSE: self._response.schema,
            OpenAPIKeyWord.EXCEPTIONS: self._response.exceptions
        })

        return result

    def _gen_call(self):
        def f(**kwargs):
            path_params = {}
            params = {}
            headers = {}
            cookies = {}

            logger.debug(
                    f"{self.operation_id} called " +
                    f"with {kwargs} arguments")
            # Set operation specific parameters
            for param in self._params:
                param.serialize(path_params, params, headers, cookies, kwargs)

            if self._body:
                self._body.serialize(headers, kwargs)

            # collect internal params
            for k in kwargs:
                if not k.startswith(self._internal_param_prefix):
                    continue
                kwargs[
                    k[len(self._internal_param_prefix) :]  # noqa: E203
                ] = kwargs.pop(k)

            kwargs.setdefault("params", {}).update(params)
            kwargs.setdefault("headers", {}).update(headers)
            kwargs.setdefault("cookies", {}).update(cookies)

            return self._session.request(
                self._method, self.url(**path_params), **kwargs
            )

        return f

    def __call__(self, *args, **kwargs):
        if not self._call:
            self._call = self._gen_call()
        return self._call(*args, **kwargs)

    def __repr__(self):
        return f"<{type(self).__name__}: [{self._method}] {self._path}>"


class Client(object):
    def __init__(self,
                 name,
                 version=None,
                 session=None,
                 residency=None,
                 variables=None):
        self._name = name
        self._server = None
        self._session = session
        self._residency = residency
        self._operations = {}
        self._spec = {}
        self._models = {}
        self._info = {}
        self.load_service_spec(name, version, variables)

    def load_service_spec(self, service_name, version=None, variables=None):
        logger.debug(
            f"Initializing client for '{self._name}' " +
            f"Spec: '{service_name}' Variables: '{variables}'")
        spec = alsdkdefs.load_service_spec(service_name, Config.get_api_dir(), version)
        self.load_spec(spec, variables)

    @property
    def name(self):
        return self._name

    @property
    def server(self):
        return self._server

    def set_server(self, s):
        self._server = s
        self._initialize_operations()

    @property
    def info(self):
        return self._info

    @property
    def description(self):
        return self._info.get(OpenAPIKeyWord.DESCRIPTION, "")

    @property
    def operations(self):
        return self._operations

    @property
    def spec(self):
        return self._spec

    def load_spec(self, spec, variables):
        if not all(
            [
                i in spec
                for i in [
                    OpenAPIKeyWord.OPENAPI,
                    OpenAPIKeyWord.INFO,
                    OpenAPIKeyWord.PATHS,
                ]
            ]
        ):
            raise ValueError("Invalid OpenAPI document")

        self._spec = spec.copy()
        _spec = spec.copy()

        self._info = _spec.pop(OpenAPIKeyWord.INFO)

        servers = _spec.pop(OpenAPIKeyWord.SERVERS, [])
        for key in _spec:
            rkey = key.replace("-", "_")
            self.__setattr__(rkey, _spec[key])

        self.servers = [
            Server(
                service_name=self._name,
                spec=s,
                session=self._session,
                variables=variables)
            for s in servers
        ]

        if not self._server and self.servers:
            if self._session:
                self._server = list(
                        filter(lambda s: self._session.validate_server(s.spec),
                               self.servers))[0]
            else:
                self._server = self.servers[0]

        self._initialize_operations()

    def _initialize_operations(self):
        self._operations = {}
        for path, path_spec in self.paths.items():
            for method, op_spec in path_spec.items():
                operation_id = op_spec.get(OpenAPIKeyWord.OPERATION_ID)
                summary = op_spec.pop(OpenAPIKeyWord.SUMMARY, "")
                description = op_spec.pop(OpenAPIKeyWord.DESCRIPTION, "")

                if not operation_id:
                    logging.warn(
                        f"'{OpenAPIKeyWord.OPERATION_ID}' not found in: \
                          '[{method}] {path}'"
                    )
                    continue

                if operation_id in self._operations:
                    raise AlmdrlibValueError(f"Duplication {operation_id} \
                                       specified for {self._name} API")

                # Initialize parameters (path, header, query)
                params = [
                    PathParameter(spec=s, session=self._session)
                    for s in op_spec.get(OpenAPIKeyWord.PARAMETERS, [])
                ]

                # Initialize operation's body
                body = self._initalize_request_body(
                        op_spec.pop(OpenAPIKeyWord.REQUEST_BODY, None)
                    )

                # Initialize operation's response
                response = OperationResponse(
                        op_spec.pop(OpenAPIKeyWord.RESPONSES, None)
                    )

                self._operations[operation_id] = Operation(
                    path,
                    params,
                    summary,
                    description,
                    method,
                    op_spec,
                    body,
                    response,
                    session=self._session,
                    server=self._server
                )

    def _initalize_request_body(self, body_spec=None):
        ''' Initilize request body content & parameters'''
        if not body_spec:
            return None

        request_body = RequestBody(
                required=body_spec.pop(OpenAPIKeyWord.REQUIRED, False),
                description=body_spec.pop(OpenAPIKeyWord.DESCRIPTION, None),
                session=self._session)

        content = body_spec.pop(OpenAPIKeyWord.CONTENT, {})
        for content_type, content_schema in content.items():
            request_body.add_content(
                    content_type,
                    content_schema.pop(OpenAPIKeyWord.SCHEMA),
                    content_schema.pop(OpenAPIKeyWord.X_ALERTLOGIC_SCHEMA, {})
            )
        return request_body

    def __getattr__(self, op_name):
        if op_name in self._operations:
            return self._operations[op_name]
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{op_name}'"
        )


def _normalize_schema(name, schema, required=False):
    properties = schema.get(OpenAPIKeyWord.PROPERTIES)
    if properties and bool(properties):
        return schema

    result = {
        OpenAPIKeyWord.TYPE: OpenAPIKeyWord.OBJECT,
        OpenAPIKeyWord.PROPERTIES: {
            name: schema
        }
    }

    if required:
        result[OpenAPIKeyWord.REQUIRED] = [name]
    return result


def get_dict_value(dict, list, default=None):
    length = len(list)
    try:
        for depth, key in enumerate(list):
            if depth == length - 1:
                output = dict[key]
                return output
            dict = dict[key]
    except (KeyError, TypeError):
        return default
    return default


def update_dict_no_replace(target, source):
    for key in source.keys():
        if key not in target:
            target[key] = source[key]


def serialize_value(datatype, value):
    if OpenAPIKeyWord.STRING == datatype:
        return value
    elif OpenAPIKeyWord.BOOLEAN == datatype:
        return value and "true" or "false"
    else:
        return str(value)



'''Cloud stored environment configuration
AlEnv class implements retrieval and formatting of configuration parameters from dynamodb in the standardized way.
AWS client used is configured as follows https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html
Therefore, AWS cli configuration would be automatically picked up.
The simpliest way is to use environment variables.
Alert Logic specific configuration is done via ALERTLOGIC_STACK_REGION and ALERTLOGIC_STACK_NAME env variables.

Exceptions:
AlEnvException is a generic type for any AlEnv specific errors, derivatives are:
AlEnvConfigurationTableUnavailableException() - thrown when configuration table is not found
AlEnvAwsConfigurationException() - AWS credentials or configuration issue

Network or io errors are not handled.

Usage:
# export ALERTLOGIC_STACK_REGION=us-east-1
# export ALERTLOGIC_STACK_NAME=integration
>> env = AlEnv("myapplication")
# Assuming parameter is stored in ddb as '"value"'
>> env.get("my_parameter")
'value'
>> env.get("my_parameter", format='raw')
'"value"'
# Assuming parameter is stored in ddb as '"1.0"'
>> env.get("my_parameter", type='float')
1.0
>> env.get("my_parameter", type='integer')
1
# Assuming parameter is stored in ddb as '"true"'
>> env.get("my_parameter", type='boolean')
True
'''

import json
import boto3
import os
import botocore


class AlEnvException(Exception):
    pass


class AlEnvAwsConfigurationException(AlEnvException):
    pass


class AlEnvConfigurationTableUnavailableException(AlEnvException):
    pass


class AlEnv:
    def __init__(self, application_name):
        self.application_name = application_name
        self.region = AlEnv._get_region()
        self.stack_name = AlEnv._get_stack_name()
        self.table_name = AlEnv._table_name(self.region, self.stack_name)
        try:
            self.dynamodb = boto3.resource('dynamodb')
            self.table = self.dynamodb.Table(self.table_name)
            self._table_date_time = self.table.creation_date_time
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                raise AlEnvConfigurationTableUnavailableException(self.table_name)
            else:
                raise AlEnvException(e)
        except (botocore.exceptions.NoRegionError, botocore.exceptions.NoCredentialsError) as e:
            raise AlEnvAwsConfigurationException(f'Please validate your AWS configuration: {e}')

    def get(self, key, default=None, format='decoded', type=None):
        fetched_value = self.table.get_item(Key={"key": self._make_ddb_key(key)}).get('Item', {}).get('value')
        converted = AlEnv._convert(fetched_value, format, type)
        if converted is not None:
            return converted
        else:
            return default

    def _make_ddb_key(self, option_key):
        return f"{self.application_name}.{option_key}"

    @staticmethod
    def _convert(value, format, type):
        if format == 'raw' and value:
            return AlEnv._format_value(value, type)
        elif format == 'decoded' and value:
            decoded = json.loads(value)
            return AlEnv._format_value(decoded, type)
        else:
            return None

    @staticmethod
    def _format_value(value, type):
        if type == 'integer':
            return int(value)
        elif type == 'float':
            return float(value)
        elif type in ['boolean', 'bool']:
            return AlEnv._to_bool(value)
        else:
            return value

    @staticmethod
    def _to_bool(value):
        if isinstance(value, bool):
            return value
        elif value == 'true':
            return True
        elif value == 'false':
            return False
        else:
            raise ValueError('Provided value cannot be converted to boolean')

    @staticmethod
    def _table_name(region, stack_name):
        global_app = "global"
        config_table = "settings"
        environment_name = 'dev'
        return f"{region}.{stack_name}.{environment_name}.{global_app}.{config_table}"

    @staticmethod
    def _get_region():
        return os.environ.get('ALERTLOGIC_STACK_REGION', 'us-east-1')

    @staticmethod
    def _get_stack_name():
        return os.environ.get('ALERTLOGIC_STACK_NAME', 'integration')



import os

from almdrlib.session import Session
from almdrlib.docs.service import ServiceDocGenerator


def make_documentation(dir, session=None):
    """Creates documentation for Alert Logic SDK

    Enumerates all services and generates rst files for sphinx documentation

    :param dir: The directory to write the files to. Each
        service's documentation is loacated at
        dir/services/[service name].rst file

    :param session: The IWS Services session
    """

    doc_path = os.path.join(dir, 'site', 'services')
    if not os.path.exists(doc_path):
        os.makedirs(doc_path)

    for name in Session.list_services():
        doc = ServiceDocGenerator(
                service_name=name,
                spec=Session.get_service_api(name)
            ).get_documentation()

        service_doc_path = os.path.join(
            doc_path, name + '.rst')
        with open(service_doc_path, 'w+') as f:
            f.writelines(f"{line}\n" for line in doc)



import unittest
from unittest.mock import MagicMock
from almdrlib.environment import AlEnv
import os
import boto3


class MockDDBTable:
    creation_date_time = '2005-08-09T18:31:42-03:30'
    def __init__(self, tablename):
        pass

    def get_item(self, **kwargs):
        Key = kwargs.get('Key')['key']
        if Key == 'someapplication.strkey':
            return {"Item": {"key": Key, "value": "\"strvalue\""}}
        elif Key == 'someapplication.intkey':
            return {"Item": {"key": Key, "value": "\"1\""}}
        elif Key == 'someapplication.floatkey':
            return {"Item": {"key": Key, "value": "\"1.0\""}}
        elif Key == 'someapplication.boolkeytrue':
            return {"Item": {"key": Key, "value": "\"true\""}}
        elif Key == 'someapplication.boolkeyfalse':
            return {"Item": {"key": Key, "value": "\"false\""}}


class MockBotoDDB:
    Table = MockDDBTable

    def __init__(self, client_type):
        pass


class TestAlEnv(unittest.TestCase):
    def test_something(self):
        boto3.resource = MagicMock(return_value=MockBotoDDB)
        os.environ['ALERTLOGIC_STACK_REGION'] = 'us-west-1'
        os.environ['ALERTLOGIC_STACK_NAME'] = 'production'
        env = AlEnv("someapplication")
        assert env.table_name == 'us-west-1.production.dev.global.settings'
        assert env.get("strkey") == 'strvalue'
        assert env.get("strkey", format='raw') == '"strvalue"'
        assert env.get("intkey") == '1'
        assert env.get("intkey", type='float') == 1.0
        assert env.get("boolkeyfalse") == 'false'
        assert env.get("boolkeyfalse", type='boolean') is False
        assert env.get("boolkeytrue", type='boolean') is True
        assert env.get("floatkey") == '1.0'
        assert env.get("floatkey", type='float') == 1.0


if __name__ == '__main__':
    unittest.main()



import json
import logging
import itertools
try:
    from m2r import convert
except Exception:
    def convert(text):
        return text

from alsdkdefs import OpenAPIKeyWord

logger = logging.getLogger(__name__)


class ServiceDocGenerator(object):
    def __init__(self,
                 service_name,
                 spec,
                 initial_indent='  ',
                 indent_increment=2):
        self._initial_indent = initial_indent
        self._indent_increment = indent_increment
        self._service_name = service_name
        self._spec = spec

    def _indent(self, indent, count=1):
        return indent + self._initial_indent * count

    def get_documentation(self):
        generators = []
        generators.append(self._make_header())
        generators.append(self._make_class())
        generators.append(self._make_methods())
        return iter(itertools.chain(*generators))

    def _make_header(self):
        yield self._service_name.capitalize()
        yield '*' * len(self._service_name)
        yield ''

        yield '.. contents:: Table of Contents'
        yield '   :depth: 2'
        yield ''

    def _make_class(self):
        yield '======'
        yield 'Client'
        yield '======'

        indent = self._initial_indent
        c_name = self._service_name.capitalize()
        yield f'.. py:class:: {c_name}.Client'
        yield ''
        yield f"{indent}A client object representing '{c_name}' Service::"
        yield ''

        yield f'{indent*2}import almdrlib'
        yield ''
        yield f"{indent*2}client = almdrlib.client('{self._service_name}')"
        yield ''

        yield f'{indent}Available methods:'
        for op_name in self._spec['operations'].keys():
            yield ''
            yield f'{indent}*   :py:meth:`~{c_name}.Client.{op_name}`'
            yield ''

    def _make_methods(self):
        indent = self._initial_indent
        for op_name, op_spec in self._spec['operations'].items():
            yield ''
            yield f'{indent}.. py:method:: {op_name}(**kwargs)'
            for line in convert(op_spec.get('description', "")).splitlines():
                yield f'{indent}{indent}{line}'

            parameters = op_spec.get('parameters', {})
            for line in self._make_request_syntax(
                    op_name, parameters, self._indent(indent)):
                yield line

            yield from self._make_request_parameters(
                    op_name, parameters, indent)

            yield from self._make_response(
                    op_spec.get('response', {}), indent)

    def _make_request_syntax(self, name, parameters, indent):
        yield ''
        yield f'{indent}**Request Syntax**'
        yield f'{indent}::'
        yield ''

        indent = self._indent(indent)
        yield f'{indent}response = client.{name}('

        param_indent = self._indent(indent, 2)
        request_spec = [
                f"{param_indent}{param_name}=" +
                format_json(get_param_spec(param_spec), param_indent)
                for param_name, param_spec in parameters.items()
            ]

        result = '\n'.join(request_spec).replace('"', "'").replace("''", "'")
        for line in result.splitlines():
            yield f'{line}'
        yield f'{indent})'

    def _make_request_parameters(self, op_name, parameters, indent):
        indent = self._indent(indent)
        for name, spec in sorted(parameters.items()):
            yield ''
            if 'content' in spec:
                type = ' | '.join([
                        get_param_type(v)
                        for v in spec['content'].values()
                    ])
            else:
                type = get_param_type(spec)

            yield f'{indent}:type {name}: {type}'
            if 'required' in spec or 'x-alertlogic-required' in spec:
                yield f'{indent}:param {name}:  **[REQUIRED]**'
            else:
                yield f'{indent}:param {name}:'
            yield ''

            for line in convert(spec.get('description', "")).splitlines():
                yield f'{indent}{indent}{line}'

            yield from self._make_property(
                    spec, name=name,
                    declare=False, indent=self._indent(indent))

            if name == 'content_type':
                yield from self._make_content_type(
                        op_name,
                        spec.pop('x-alertlogic-payload-content'),
                        indent=indent)

            elif 'content' in spec:
                yield from self._make_request_body_parameter(
                        spec.pop('content'),
                        name,
                        indent=self._indent(indent)
                    )
            else:
                yield from self._make_properties(
                        spec=spec,
                        indent=self._indent(indent),
                        declare=False
                    )

    def _make_content_type(self, op_name, spec, indent):
        if not spec:
            return

        content_indent = self._indent(indent)
        valid_values = ', '.join([
                f'``{v}``'
                for v in spec.keys()]
            )
        yield ''
        yield f'{content_indent}*Valid values*: {valid_values}'
        yield ''
        yield f'{content_indent}.. note::'
        content_indent = self._indent(content_indent)
        yield (f'{content_indent}Following parameters depend on the'
               '``content_type`` value:')

        properties = []
        for property in itertools.chain.from_iterable(spec.values()):
            if property not in properties:
                properties.append(property)

        prop_links = [
                f':paramref:`.{op_name}.{v}`'
                for v in sorted(properties)
            ]
        yield f'{content_indent}{", ".join(prop_links)}'

    def _make_request_body_parameter(self, specs, name, indent=None):
        yield ''
        yield f'{indent}Below is specification of ``{name}`` parameter '
        yield (f"{indent}as it relates to the ``content_type`` "
               "parameter's value")

        for content_type, spec in specs.items():
            yield ''
            spec_indent = self._indent(indent)
            yield f'{spec_indent}.. toggle-header::'

            title = f'``content_type`` == {content_type}'
            spec_indent = self._indent(spec_indent)
            yield f'{spec_indent}:header: {title}'

            yield ''
            spec_indent = self._indent(spec_indent)
            yield from self._make_property(spec, indent=spec_indent)
            yield from self._make_properties(
                    spec=spec,
                    indent=self._indent(spec_indent))

    def _make_properties(self, spec, name=None, indent=None, declare=True):
        type = get_param_type(spec)

        if type == 'list':
            if 'items' in spec:
                yield from self._make_property(
                        spec['items'], name=name, indent=indent)
                yield from self._make_properties(
                        spec=spec['items'], indent=self._indent(indent))

            elif 'enum' in spec:
                yield from self._make_property(spec, indent=indent)

            else:
                logger.warn(
                        f"Unknown key for data type for {name}: {type}."
                        f"{json.dumps(spec, indent=2)}")

        elif type == 'dict':
            for p_name, p_spec in sorted(spec.get('properties', {}).items()):
                yield from self._make_property(
                        spec=p_spec, name=p_name, indent=indent)
                yield from self._make_properties(
                        spec=p_spec,
                        indent=self._indent(indent))

        elif type == 'oneOf' or type == 'anyOf':
            yield from self._make_indirect_property(
                    specs=spec.get('oneOf', spec.get('anyOf')),
                    name=name, indent=indent)

    def _make_indirect_property(self, specs, name=None, indent=None):
        yield ''
        yield f'{indent}.. content-tabs::'
        yield ''
        for spec in specs:
            title = spec.get('title', spec.get('type', name))
            spec_indent = self._indent(indent)
            yield ''
            yield f'{spec_indent}.. tab-container:: {title}'

            spec_indent = self._indent(spec_indent)
            yield f'{spec_indent}:title: {title}'
            yield from self._make_property(spec, indent=spec_indent)
            yield from self._make_properties(
                    spec=spec,
                    indent=spec_indent)

    def _make_property(self, spec, name=None, indent=None, declare=True):
        yield ''
        if declare:
            type = get_param_type(spec)
            if name:
                if 'required' in spec or 'x-alertlogic-required' in spec:
                    yield f'{indent}- **{name}** *({type}) --* **[REQUIRED]**'
                else:
                    yield f'{indent}- **{name}** *({type}) --*'
            else:
                yield f'{indent}- *({type}) --*'

        indent = self._indent(indent)
        if declare:
            yield ''
            for line in convert(spec.get('description', "")).splitlines():
                yield f'{indent}{line}'

        if 'enum' in spec:
            valid_values = ', '.join([f'``{v}``' for v in spec['enum']])
            yield ''
            yield f'{indent}*Valid values*: {valid_values}'

        if 'default' in spec:
            yield ''
            yield f'{indent}*Default*: ``{spec["default"]}``'

    def _make_response(self, spec, indent):
        yield ''
        if not bool(spec):
            yield f'{indent}:returns: None'
            yield ''
            return ''

        yield f'{indent}:rtype: {get_param_type(spec)}'
        yield f'{indent}:returns:'

        indent = self._indent(indent)
        yield f'{indent}**Response Syntax**'
        yield ''
        yield f'{indent}::'

        yield ''
        syntax_indent = self._indent(indent)
        syntax = format_json(get_param_spec(spec))
        for line in syntax.replace('"', "'").replace("''", "'").splitlines():
            yield f'{syntax_indent}{line}'

        yield ''
        yield f'{indent}**Response Definitions**'
        yield ''

        yield from self._make_properties(
                spec=spec,
                indent=self._indent(indent))


def get_param_spec(spec):
    if OpenAPIKeyWord.ONE_OF in spec:
        types = [get_param_type(v) for v in spec[OpenAPIKeyWord.ONE_OF]]
        return '|'.join(list(set(sorted(types))))

    if OpenAPIKeyWord.ANY_OF in spec:
        types = [get_param_type(v) for v in spec[OpenAPIKeyWord.ANY_OF]]
        return '|'.join(list(set(sorted(types))))

    if OpenAPIKeyWord.CONTENT in spec:
        content = spec.get(OpenAPIKeyWord.CONTENT, {})
        return '|'.join(
                sorted([get_param_type(v) for v in content.values()])
            )

    param_type = spec.get(OpenAPIKeyWord.TYPE)
    if param_type == OpenAPIKeyWord.ARRAY:
        return [get_param_spec(spec.get(OpenAPIKeyWord.ITEMS, {}))]

    if param_type == OpenAPIKeyWord.OBJECT:
        properties = spec.get(OpenAPIKeyWord.PROPERTIES, {})
        return {k: get_param_spec(v) for k, v in sorted(properties.items())}

    if param_type == OpenAPIKeyWord.BOOLEAN:
        return "False|True"

    enum = spec.get(OpenAPIKeyWord.ENUM)
    if enum:
        return '|'.join(f"'{v}'" for v in enum)
    else:
        return f"'{param_type}'"


def get_param_type(spec):
    type = spec.get('type', 'object')
    format = spec.get('format')

    if type == 'object':
        return 'dict'
    elif type == 'array':
        return 'list'
    elif format:
        return format
    elif type:
        return type
    else:
        if 'oneOf' in spec:
            return 'oneOf'
        elif 'anyOf' in spec:
            return 'anyOf'
        else:
            raise ValueError(
                f'Unsupported parameter type. {json.dumps(spec, indent=2)}')


def format_json(value, indent=None):
    if not indent:
        indent = ''
    res = json.dumps(value, sort_keys=True, indent=4).splitlines()
    return f'\n{indent}'.join(res)



# -*- coding: utf-8 -*-

"""
    almdrlib.session
    ~~~~~~~~~~~~~~~~
    almdrlib authentication/authorization
"""
import types
import os
import logging

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from almdrlib.config import Config
from almdrlib.region import Region
from almdrlib.client import Client
import alsdkdefs

logger = logging.getLogger(__name__)


class AuthenticationException(Exception):
    def __init__(self, message):
        super(AuthenticationException, self).__init__(
                f"authentication error: {message}")


class Session():
    """
    Authenticates against Alert Logic AIMS service and
    stores session information (token and account id).
    Additionally objects of this class can be used as auth modules
    for the requests lib, more info:
    http://docs.python-requests.org/en/master/user/authentication/#new-forms-of-authentication
    """

    _access_key_id = None
    _secret_key = None

    def __init__(
            self, access_key_id=None, secret_key=None, aims_token=None,
            account_id=None, profile=None, global_endpoint=None,
            residency="default", raise_for_status=True):
        """
        :param region: a Region object
        :param access_key_id: your Alert Logic ActiveWatchaccess_key_id
                              or username
        :param secret_key: your Alert Logic ActiveWatchsecret_key or password
        :param aims_token: aims_token to be used for authentication.
                           If aims_token is specified,
                           access_key_id and secret_key paramters are ignored
        :param account_id: Alert Logic Account ID to initialize a session for.
                            Unless account_id is provided explicitly
                            during service connection initialization,
                            this account id is used.
                            If this parameter isn't specified,
                            the account id of the access_key_id is used.
        :param: profile: name of the profile section of the configuration file
        :param: global_endpoint: Name of the global endpoint.
                                 'production', 'integration', or 'map' are
                                 the only valid values
        :param residency: Data residency name to perform
                          data residency dependend actions.
                          Currently, 'default', 'us' and 'emea'
                          are the only valid entries
        :param raise_for_status: Raise an exception for failed http requests
                                 instead of returning response object
        """

        # Setup session object
        self._session = requests.Session()
        retries = Retry(
                total=5,
                backoff_factor=1,
                status_forcelist=[429, 500, 502, 503, 504],
                method_whitelist=[
                    "HEAD", "GET", "POST", "PUT", "DELETE", "OPTIONS", "TRACE"]
                )
        self._session.mount('https://', HTTPAdapter(max_retries=retries))

        # Initialize session's state
        self._token = None
        self._defaults = None
        self._init_session(
                access_key_id=access_key_id,
                secret_key=secret_key,
                aims_token=aims_token,
                account_id=account_id,
                profile=profile,
                global_endpoint=global_endpoint,
                residency=residency,
                raise_for_status=raise_for_status)

    def _init_session(self, *args, **kwargs):
        """ Initialize session object based on the kwargs provided """

        if not len(kwargs):
            return

        access_key_id = kwargs.get('access_key_id')
        secret_key = kwargs.get('secret_key')
        account_id = kwargs.get('account_id')
        profile = kwargs.get('profile')
        global_endpoint = kwargs.get('global_endpoint')
        residency = kwargs.get('residency', 'default')
        aims_token = kwargs.get('aims_token')

        self._config = Config(
                access_key_id=access_key_id,
                secret_key=secret_key,
                account_id=account_id,
                profile=profile,
                global_endpoint=global_endpoint,
                residency=residency
            )

        self._account_id = self._config.account_id
        self._residency = self._config.residency
        self._global_endpoint = self._config.global_endpoint
        self._endpoint_map = self._config.endpoint_map
        self._global_endpoint_url = Region.get_global_endpoint(
                                                        self._global_endpoint)
        self._raise_for_status = kwargs.get('raise_for_status')

        if aims_token:
            self._token = aims_token
        else:
            self._access_key_id, self._secret_key = self._config.get_auth()

        logger.debug(
                "Initialized session. "
                f"access_key_id={self._access_key_id}, "
                f"account_id={self._account_id}, "
                f"profile={profile}, "
                f"global_endpoint={self._global_endpoint}, "
                f"residency={self._residency}"
            )

    def _authenticate(self):
        """
        Authenticates against Access and Identity Management Service (AIMS)
        more info:
        https://console.cloudinsight.alertlogic.com/api/aims/#api-AIMS_Authentication_and_Authorization_Resources-Authenticate
        """

        if not self._token:
            if self._access_key_id == "skip" and self._secret_key == "skip":
                logger.info(
                        f"Skipping authentication."
                    )
                self._token = ""
                self._account_id = ""
                self._account_name = ""
                self._user_id = ""
                return
            logger.info(
                    f"Authenticating '{self._access_key_id}' " +
                    f"user against '{self._global_endpoint_url}' endpoint."
                )
            try:
                self._session.auth = (self._access_key_id, self._secret_key)
                response = self._session.post(
                        f"{self._global_endpoint_url}/aims/v1/authenticate")
                response.raise_for_status()

                auth_info = response.json()
                account_info = auth_info["authentication"]["account"]
                self._token = auth_info["authentication"]["token"]
                self._user_id = auth_info["authentication"]["user"]["id"]
                logger.info(f'Authenticated user {auth_info["authentication"]["user"]["id"]}')

            except requests.exceptions.HTTPError as e:
                raise AuthenticationException(f"invalid http response {e}")
            except (KeyError, TypeError, ValueError):
                raise AuthenticationException("token not found in response")
        else:
            logger.info("Authenticating using aims token " +
                        f"against '{self._global_endpoint_url}' endpoint.")
            try:
                response = self._session.get(
                        f"{self._global_endpoint_url}/aims/v1/token_info",
                        headers={'x-aims-auth-token': self._token})
                response.raise_for_status()
                account_info = response.json()["account"]

            except requests.exceptions.HTTPError:
                self._token = None
                return self._authenticate()
            except (KeyError, TypeError, ValueError):
                raise AuthenticationException(
                        "account information not found in response")

        if self._account_id is None:
            try:
                self._account_id = account_info["id"]
                self._account_name = account_info["name"]
            except (KeyError, TypeError, ValueError):
                raise AuthenticationException(
                        "account information not found in response")

    def __call__(self, r):
        """
        requests lib auth module callback
        """
        if self._token is None:
            self._authenticate()
        r.headers["x-aims-auth-token"] = self._token
        return r

    def client(self, service_name, version=None, *args, **kwargs):
        """
        Create Service's client class
        """

        self._init_session(**kwargs)

        # Create Service's module
        module_name = service_name.capitalize()
        types.ModuleType = module_name
        class_name = "Client"

        #
        # Init function for the dynamically created class,
        # which is derived from almdrlib.client.Client
        #
        def __init__(self,
                     name,
                     session=self,
                     version=None,
                     *args,
                     **kwargs):
            super(self.__class__, self).__init__(name=name,
                                                 session=session,
                                                 version=version)

        ServiceClient = type(class_name,
                             (Client,),
                             {
                                 '__init__': __init__,
                                 '__module__': module_name
                             })

        _client = ServiceClient(service_name, session=self, version=version)
        logger.debug(
            "Created " +
            f"{_client.__class__.__module__}.{_client.__class__.__name__}" +
            " class instance")
        return _client

    def get_url(self, service_name, account_id=None):
        if self._global_endpoint == "map":
            return self.get_mapped_url(service_name, account_id)
        try:
            response = self.request(
                'get',
                Region.get_endpoint_url(self._global_endpoint_url,
                                        service_name,
                                        account_id or self.account_id,
                                        self.residency),
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise AuthenticationException(
                    f"invalid http response from endpoints service {e}"
                )
        return "https://{}".format(response.json()[service_name])

    def get_mapped_url(self, service_name, account_id):
        map = self._endpoint_map
        return map[service_name]

    def request(
            self,
            method,
            url,
            params={},
            headers={},
            cookies={},
            **kwargs):

        # Make sure we've authenticated
        if self._token is None:
            self._authenticate()

        # it's too easy to include the AIMS token when pasting debug logs, so redact it in
        # the logging statement.
        headers.update({'x-aims-auth-token': "REDACTED"})

        logger.debug(f"Calling '{method}' method. " +
                     f"URL: '{url}'. " +
                     f"Params: '{params}' " +
                     f"Headers: '{headers}' " +
                     f"Cookies: '{cookies}' " +
                     f"Args: '{kwargs}'")

        headers.update({'x-aims-auth-token': self._token})

        response = self._session.request(
                method, url,
                params=params,
                headers=headers,
                cookies=cookies,
                **kwargs)
        if self._raise_for_status:
            response.raise_for_status()
        logger.debug(f"'{method}' method for URL: '{url}' returned "
                     f"'{response.status_code}' status code "
                     f"in '{response.elapsed.total_seconds()}' seconds")
        return response

    def get_default(self, name):
        if name == 'account_id':
            return self.account_id

        return None

    def validate_server(self, spec):
        if 'x-alertlogic-global-endpoint' not in spec:
            return True

        return self._global_endpoint == spec['x-alertlogic-global-endpoint']

    @staticmethod
    def list_services():
        return alsdkdefs.list_services()

    @staticmethod
    def get_service_api(service_name, version=None):
        client = Client(service_name, version=version)
        model = {'info': client.info}
        operations = {}
        for op_name, operation in client.operations.items():
            operations[op_name] = operation.get_schema()
        model.update({'operations': dict(sorted(operations.items()))})
        return model

    @property
    def account_id(self):
        if self._account_id is None:
            self._authenticate()
        return self._account_id

    @property
    def residency(self):
        return self._residency

    @property
    def account_name(self):
        return self._account_name

    @property
    def global_endpoint(self):
        return self._global_endpoint

    @property
    def global_endpoint_url(self):
        return self._global_endpoint_url

    @property
    def token(self):
        return self._token

    @property
    def user_id(self):
        return self._user_id



__author__ = 'Alert Logic, Inc.'

import logging
import almdrlib.constants
from almdrlib.config import Config
from almdrlib.session import Session

IWS_DEFAULT_SESSION = None


def set_logger(name='almdrlib', level=logging.DEBUG, format_string=None):
    """
    Add a stream handler for the given name and level to the logging module.
    By default, this logs all almdrlib messages to ``stdout``.

        >>> import almdrlib
        >>> almdrlib.set_logger('almdrlib.client', logging.INFO)

    For debugging purposes a good choice is to set the stream logger to ``''``
    which is equivalent to saying "log everything".

    :type name: string
    :param name: Log name
    :type level: int
    :param level: Logging level, e.g. ``logging.INFO``
    :type format_string: str
    :param format_string: Log message format
    """

    if format_string is None:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)

    # avoid duplicate logs if already set by caller, etc.
    if not logger.handlers:
        logger.addHandler(handler)


def _get_default_session():
    global IWS_DEFAULT_SESSION
    if not IWS_DEFAULT_SESSION:
        IWS_DEFAULT_SESSION = Session()
    return IWS_DEFAULT_SESSION


def client(service_name, version=None, session=None, *args, **kwargs):
    if session is None:
        session = _get_default_session()
    return session.client(service_name, version, *args, **kwargs)


def configure(
        profile=almdrlib.constants.DEFAULT_PROFILE,
        access_key_id=None, secret_key=None,
        global_endpoint=almdrlib.constants.DEFAULT_GLOBAL_ENDPOINT,
        residency=almdrlib.constants.DEFAULT_RESIDENCY):
    return Config.configure(
            profile=profile,
            access_key_id=access_key_id, secret_key=secret_key,
            global_endpoint=global_endpoint, residency=residency)


# Logging to dev/null
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


logging.getLogger('almdrlib').addHandler(NullHandler())



# -*- coding: utf-8 -*-

import os.path
import configparser
import logging
import json
import almdrlib.constants
from almdrlib.exceptions import AlmdrlibValueError
from almdrlib.environment import AlEnv

logger = logging.getLogger(__name__)


class ConfigException(Exception):
    def __init__(self, message):
        super(ConfigException, self).__init__("config error: {message}")


class Config():
    """
    Reads configuration parameters from either environment variables
        or from configuration file

    Environment Variables:
    ALERTLOGIC_CONFIG - Location of the configuration file. If not specified,
    ~/.alertlogic/credentials is used
    ALERTLOGIC_PROFILE - Profile to be used.
    If not specified [default] section is used
    ALERTLOGIC_ACCESS_KEY_ID - Access Key Id
    ALERTLOGIC_SECRET_KEY - Secret Key
    ALERTLOGIC_ENDPOINT - production, integration, or map are supported values
    ALERTLOGIC_ENDPOINT_MAP - (full or relative to config directory) path of a json file mapping services to endpoints
    ALERTLOGIC_ACCOUNT_ID - Account Id to perform operations against.
    ALERTLOGIC_RESIDENCY - Data Residency when creating new deployments
    ALERTLOGIC_API  - Directory where OpenAPI yaml files reside
    ALERTLOGIC_SERVICE_NAME - If a (micro)service is built around almdrlib used as identifier

    Config File section values
    access_key_id - User's AIMS Access Key ID
    secret_key - Secret Key associated with 'access_key_id'
    global_endpoint - if not specified, 'production' endpoint is used
    endpoint_map_file - if not specified, 'endpoint_map.json' in the config directory (default: ~/.alertlogic) is used
    account_id - if not specified, the account id of the access_key_id is used
    residency - if not specified, 'us' residency is used

    NOTE: Environment variables take precedence over values
          specified in configuration file
    """
    def __init__(self,
                 access_key_id=None,
                 secret_key=None,
                 account_id=None,
                 profile=None,
                 global_endpoint=None,
                 endpoint_map_file=None,
                 residency=None,
                 service_name=None):
        self._config_file = os.environ.get('ALERTLOGIC_CONFIG')
        self._endpoint_map = None
        self._service_name = service_name

        if self._config_file is None:
            self._config_file = almdrlib.constants.DEFAULT_CONFIG_FILE

        logger.debug(
                "Initializing configuration using " +
                f"'{self._config_file}' configuration file")
        if access_key_id or secret_key:
            self._access_key_id = access_key_id
            self._secret_key = secret_key
        else:
            self._access_key_id = os.environ.get('ALERTLOGIC_ACCESS_KEY_ID')
            self._secret_key = os.environ.get('ALERTLOGIC_SECRET_KEY')

        self._global_endpoint = \
            global_endpoint or \
            os.environ.get('ALERTLOGIC_ENDPOINT')

        self._endpoint_map_file = \
            endpoint_map_file or \
            os.environ.get('ALERTLOGIC_ENDPOINT_MAP')

        self._residency = \
            residency or \
            os.environ.get('ALERTLOGIC_RESIDENCY')

        self._account_id = \
            account_id or \
            os.environ.get('ALERTLOGIC_ACCOUNT_ID')

        self._profile = \
            profile or \
            os.environ.get('ALERTLOGIC_PROFILE') or \
            almdrlib.constants.DEFAULT_PROFILE

        self._service_name = \
            service_name or \
            os.environ.get('ALERTLOGIC_SERVICE_NAME') or \
            almdrlib.constants.DEFAULT_SERVICE_NAME

        self._parser = configparser.ConfigParser()
        if self._read_config_file():
            self._initialize_config()
        else:
            self._initialize_defaults()

        self._init_al_env_credentials(service_name)

        logger.debug("Finished configuraiton initialization. " +
                     f"access_key_id={self._access_key_id}, " +
                     f"account_id={self._account_id}, " +
                     f"global_endpoint={self._global_endpoint}")

    def _init_al_env_credentials(self, service_name):
        try:
            if self._access_key_id is None or self._secret_key is None:
                service_name_key = f"{service_name}.aims_authc",
                env = AlEnv(service_name_key)
                self._access_key_id = env.get('access_key_id')
                self._secret_key = env.get('secret_access_key')
        except Exception as e:
            logger.debug(f"Did not initialise aims credentials for {service_name} because {e}")

    def _read_config_file(self):
        try:
            read_ok = self._parser.read(self._config_file)
            return False if self._config_file not in read_ok else True
        except configparser.MissingSectionHeaderError:
            raise ConfigException(
                    f"Invalid format in file {self._config_file}")

    def _initialize_defaults(self):
        self._global_endpoint = \
            self._global_endpoint or \
            almdrlib.constants.DEFAULT_GLOBAL_ENDPOINT

        self._residency = \
            self._residency or \
            almdrlib.constants.DEFAULT_RESIDENCY

    def _initialize_config(self):
        if self._access_key_id is None or self._secret_key is None:
            self._access_key_id = self._get_config_option('access_key_id', None)
            self._secret_key = self._get_config_option('secret_key', None)

        self._global_endpoint = \
            self._global_endpoint or self._get_config_option(
                'global_endpoint',
                almdrlib.constants.DEFAULT_GLOBAL_ENDPOINT)

        self._endpoint_map_file = \
            self._endpoint_map_file or self._get_config_option(
                'endpoint_map_file',
                almdrlib.constants.DEFAULT_ENDPOINT_MAP_FILE)
        self._endpoint_map_file = \
            self._endpoint_map_file or \
            almdrlib.constants.DEFAULT_ENDPOINT_MAP_FILE

        if self._global_endpoint == "map":
            if os.path.isabs(self._endpoint_map_file):
                path = self._endpoint_map_file
            else:
                path = os.path.join(almdrlib.constants.DEFAULT_CONFIG_DIR, self._endpoint_map_file)
            with open(path) as json_file:
                self._endpoint_map = json.load(json_file)

        self._residency = \
            self._residency or \
            self._get_config_option(
                'residency',
                almdrlib.constants.DEFAULT_RESIDENCY)

        self._account_id = \
            self._account_id or self._get_config_option('account_id', None)

    def _get_config_option(self, option_name, default_value):
        if self._parser.has_option(self._profile, option_name):
            return self._parser.get(self._profile, option_name)
        else:
            return default_value

    def get_auth(self):
        return self._access_key_id, self._secret_key

    @staticmethod
    def configure(
            profile=almdrlib.constants.DEFAULT_PROFILE,
            account_id=None,
            access_key_id=None, secret_key=None,
            global_endpoint=None,
            endpoint_map_file=None,
            residency=None):

        if not access_key_id:
            raise AlmdrlibValueError("Missing access_key_id")

        if not secret_key:
            raise AlmdrlibValueError("Missing secret_key")

        parser = _get_config_parser(almdrlib.constants.DEFAULT_CONFIG_FILE)

        try:
            parser.add_section(profile)
        except configparser.DuplicateSectionError:
            # section alread exists.
            pass
        except configparser.ValueError:
            # almdrlib.constants.DEFAULT_PROFILE was passed as the section name
            pass

        parser.set(profile, 'access_key_id', access_key_id)
        parser.set(profile, 'secret_key', secret_key)

        if account_id:
            parser.set(profile, 'account_id', account_id)

        if global_endpoint:
            parser.set(profile, 'global_endpoint', global_endpoint)

        if endpoint_map_file:
            parser.set(profile, 'endpoint_map_file', endpoint_map_file)

        if residency:
            parser.set(profile, 'residency', residency)

        with open(almdrlib.constants.DEFAULT_CONFIG_FILE, 'w') as configfile:
            parser.write(configfile)

    @staticmethod
    def set_option(
            name,
            value,
            profile=almdrlib.constants.DEFAULT_PROFILE):
        parser = _get_config_parser(almdrlib.constants.DEFAULT_CONFIG_FILE)
        parser.set(profile, name, value)
        with open(almdrlib.constants.DEFAULT_CONFIG_FILE, 'w') as configfile:
            parser.write(configfile)

    @property
    def profile(self):
        return self._profile

    @property
    def account_id(self):
        return self._account_id

    @property
    def global_endpoint(self):
        return self._global_endpoint

    @property
    def endpoint_map(self):
        return self._endpoint_map

    @property
    def residency(self):
        return self._residency

    @staticmethod
    def get_api_dir():
        api_dir = os.environ.get('ALERTLOGIC_API')
        return api_dir and f"{api_dir}"


def _get_config_parser(config_file=almdrlib.constants.DEFAULT_CONFIG_FILE):
    parser = configparser.ConfigParser()
    try:
        read_ok = parser.read(config_file)
        if config_file not in read_ok:
            raise AlmdrlibValueError(
                f"'{config_file}' doesn't exist")

    except configparser.MissingSectionHeaderError:
        raise ConfigException(
                f"Invalid format in file {config_file}")
    return parser



#!/usr/bin/env python
import os
import json

"""Tests for `alertlogic-sdk-python` package."""

import unittest

from almdrlib.session import Session
from almdrlib.client import Client
from almdrlib.client import Config
from almdrlib.client import Operation
from alsdkdefs import OpenAPIKeyWord


class TestSdk_open_api_support(unittest.TestCase):
    """Tests for `python_boilerplate` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        os.environ["ALERTLOGIC_API"] = f"{dir_path}/apis"
        self._service_name = "testapi"

        test_data_file = None
        test_data_file_path = f"{dir_path}/data/test_open_api_support.json"
        with open(test_data_file_path, "r") as value_file:
            test_data_file = value_file.read()

        self.test_data = json.loads(test_data_file)

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_getting_schemas(self):
        """Test listing services."""
        self.assertTrue(len(Session.get_service_api("testapi")))
        print(f"SCHEMA: {json.dumps(Session.get_service_api('testapi'))}")

    def test_001_test_client_initialization(self):
        """Test OpenAPI Client initialization."""
        client = Client(self._service_name)
        self.assertIsInstance(client, Client)
        self.assertIsNotNone(client.description)
        self.assertNotEqual(client.name, "")
        self.assertIsNotNone(client.server)
        self.assertNotEqual(client.info, {})
        self.assertIsNotNone(client.operations)
        self.assertNotEqual(client.spec, {})
        self.assertNotEqual(client.operations, {})

    def test_002_test_operations_schema(self):
        """Test Operations Schema Validity."""
        client = Client(self._service_name)

        for t_operation_name, t_operation_schema in \
                self.test_data['operations'].items():
            operation = client.operations.get(t_operation_name)
            self.assertIsInstance(operation, Operation)
            self.assertEqual(operation.operation_id, t_operation_name)
            self.assertEqual(operation.description,
                             t_operation_schema[OpenAPIKeyWord.DESCRIPTION])

            schema = operation.get_schema()
            self.assertIsNot(schema, {})

            t_operation_parameters = t_operation_schema[
                OpenAPIKeyWord.PARAMETERS]

            operation_parameters = schema[OpenAPIKeyWord.PARAMETERS]
            for name, value in t_operation_parameters.items():
                self.assertEqual(value, operation_parameters[name])

            if OpenAPIKeyWord.CONTENT in t_operation_schema:
                t_operation_content = t_operation_schema[
                    OpenAPIKeyWord.CONTENT]
                operation_content = schema[OpenAPIKeyWord.CONTENT]
                for name, value in t_operation_content.items():
                    self.assertEqual(value, operation_content[name])

    def test_003_default_objects_creation(self):
        """Checks initialisation at least happens"""
        self.assertIsInstance(Session(), Session)
        self.assertIsInstance(Config(), Config)
        self.assertIsInstance(Client(self._service_name), Client)



import re
import json
from urllib import request
from setuptools import setup, find_packages

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('README.md') as readme_file:
    readme = readme_file.read()

try:
    # This is to force definitions to be upgraded every build not related to the definitions change
    defs_url_pypi = "https://pypi.org/pypi/alertlogic-sdk-definitions/json"
    with request.urlopen(defs_url_pypi) as defs_rq:
        defs_info = json.loads(defs_rq.read())
    definitions_latest_version = defs_info['info']['version']
    definitions_dependency = 'alertlogic-sdk-definitions>=' + definitions_latest_version
except:
    definitions_dependency = 'alertlogic-sdk-definitions>=v0.0.74'

requirements = [
        'requests>=2.18',
        'configparser>=4.0.2',
        'pyyaml==5.1.2',
        'jsonschema[format_nongpl]==3.2.0',
        'm2r==0.2.1',
        'boto3>=1.16.57',
        definitions_dependency
    ]

test_requirements = [ ]

setup(
    name='alertlogic-sdk-python',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    url='https://github.com/alertlogic/alertlogic-sdk-python',
    license='MIT license',
    author='Alert Logic Inc.',
    author_email='devsupport@alertlogic.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    description='Alert Logic Software Development Kit for Python.',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    scripts=[],
    packages=find_packages(exclude=['contrib', 'docs', 'tests*', 'troubleshooting']),
    include_package_data=True,
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
    platforms='any',
    install_requires=requirements,
    extras_require={
        'dev': [
            'pytest>=3',
            'mock>=2.0.0',
            'httpretty>=0.8.14',
            'pycodestyle>=2.3.1',
            'jsonschema[format_nongpl]==3.2.0'
        ],
    },
    keywords=['alertlogic-sdk', 'alertlogic-sdk-python', 'alertlogic-mdr-sdk', 'almdrlib', 'alertlogic']
)



# -*- coding: utf-8 -*-

import os.path

DEFAULT_CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".alertlogic")
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_DIR, "config")
DEFAULT_CREDENTIALS_FILE = os.path.join(DEFAULT_CONFIG_DIR, "credentials")
DEFAULT_PROFILE = "default"
DEFAULT_GLOBAL_ENDPOINT = "production"
DEFAULT_ENDPOINT_MAP_FILE = "endpoint_map.json"
DEFAULT_RESIDENCY = "default"
DEFAULT_SERVICE_NAME = "almdrlib"
